/*++

Authur: Version 1.1
	Michael Stevens
	michael@mech.eng.usyd.edu.au
Previous Author: Versions to 2.0
	Jari Salminen


Module Name:

    readwrite.c

Abstract:

    Read Write processing of Irp's
    for the Transputer B008 Motherboard device driver.

Environment:

    Kernel mode

--*/

#include <ntddk.h>
#include <devioctl.h>
#include "B008drv.h"
#include "LinkIoctl.h"
#include "linksup.h"
#include "readwrite.h"


NTSTATUS
TpdWrite(
    IN    PDEVICE_OBJECT pDO,
    IN    PIRP pIrp
    )

/*++

Routine Description:
    This routine is the Write dispatch handler for the driver.  It is responsible
    for processing Write IRPs.

Arguments:

    pDO - Pointer to device object.

    pIrp - Pointer to the current IRP.

Return Value:

    STATUS_SUCCESS if the IRP was processed successfully, otherwise an error
    indicating the reason for failure.

--*/

{
    PDEVICE_INFO		pDI = (PDEVICE_INFO)pDO->DeviceExtension;
    PIO_STACK_LOCATION	pIrpStack;
	PUCHAR				pIOBuffer;		// Pointer to system buffer


	// IRP information
    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
    pIOBuffer = pIrp->AssociatedIrp.SystemBuffer;

    // Dispatch based on major fcn code.
    ASSERT (pIrpStack->MajorFunction == IRP_MJ_WRITE);

	// Illegal to write zero bytes
	if ( pIrpStack->Parameters.Write.Length < 1 ) {
	    pIrp->IoStatus.Information = 0;
	    pIrp->IoStatus.Status = STATUS_INVALID_PARAMETER;
	}

	else {
		// Output to the C012 using programed IO
		//
		ULONG 			nWriteLength = pIrpStack->Parameters.Write.Length;
		ULONG 			nWrite;
		BOOLEAN			bTimeout = FALSE;
		BOOLEAN			bInfTimeout;
		LARGE_INTEGER	liMaxTicks;
		LARGE_INTEGER	liTicks;
		LARGE_INTEGER	liTicksStart;
		PUCHAR	C012_osr = pDI->PortBase + OFFSET_OSR;
		PUCHAR	C012_odr = pDI->PortBase + OFFSET_ODR;

		// Convert nTimeout to 100 ns units, then divide by number of ticks in 100 ns to get the nTimeout in ticks
		bInfTimeout = pDI->Timeout == 0L;
		if (!bInfTimeout)
			liMaxTicks = LiXDiv (LiNMul(pDI->Timeout,10000L), KeQueryTimeIncrement() );

		// Reset timout counter
		KeQueryTickCount (&liTicksStart) ;

		// Spin, writing C012 and busy waiting for Ready Bit
		for ( nWrite = 0; nWrite < nWriteLength && !bTimeout; nWrite++ ) {
			for (;;) {
				if ( READ_PORT_UCHAR(C012_osr) & C012_OSR_READY ) {
					WRITE_PORT_UCHAR (C012_odr, *pIOBuffer++);
					// Reset timout counter
					KeQueryTickCount (&liTicksStart);
					break;
				}
				else if (!bInfTimeout) {
					// Nothing Present, check timeout
					KeQueryTickCount (&liTicks);
					liTicks.QuadPart = liTicks.QuadPart - liTicksStart.QuadPart;
					if (liTicks.QuadPart > liMaxTicks.QuadPart) {
						bTimeout = TRUE;
						break;
					}
				}
			}// Wait for Present loop
		}// Read loop

	    // Write completed, last byte wasn't written if timeout occured       
	    if (bTimeout) nWrite--;

    	pIrp->IoStatus.Information = nWrite;		// Actual number written
	    pIrp->IoStatus.Status = STATUS_SUCCESS;

	}// Valid Write

    // Don't boost priority as completed with queing Irp
    IoCompleteRequest (pIrp, IO_NO_INCREMENT);

    return pIrp->IoStatus.Status;
}


NTSTATUS
TpdRead(
    IN    PDEVICE_OBJECT pDO,
    IN    PIRP pIrp
    )

/*++

Routine Description:
    This routine is the Read dispatch handler for the driver.  It is responsible
    for processing Read IRPs.

Arguments:

    pDO - Pointer to device object.

    pIrp - Pointer to the current IRP.

Return Value:

    STATUS_SUCCESS if the IRP was processed successfully, otherwise an error
    indicating the reason for failure.

--*/
{
    PDEVICE_INFO		pDI = (PDEVICE_INFO)pDO->DeviceExtension;
    PIO_STACK_LOCATION	pIrpStack;
	PUCHAR				pIOBuffer;		// Pointer to system buffer


	// IRP information
    pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
    pIOBuffer = pIrp->AssociatedIrp.SystemBuffer;

    // Dispatch based on major fcn code.
    ASSERT (pIrpStack->MajorFunction == IRP_MJ_READ);

	// Illegal to read zero bytes
	if ( pIrpStack->Parameters.Read.Length < 1 ) {
	    pIrp->IoStatus.Information = 0;
	    pIrp->IoStatus.Status = STATUS_INVALID_PARAMETER;
	}

	else {
		// Read the C012 using programed IO
		//
		ULONG 			nReadLength = pIrpStack->Parameters.Read.Length;
		ULONG 			nRead;
		BOOLEAN			bTimeout = FALSE;
		BOOLEAN			bInfTimeout;
		LARGE_INTEGER	liMaxTicks;
		LARGE_INTEGER	liTicks;
		LARGE_INTEGER	liTicksStart;
		PUCHAR	C012_isr = pDI->PortBase + OFFSET_ISR;
		PUCHAR	C012_idr = pDI->PortBase + OFFSET_IDR;

		// Convert nTimeout to 100 ns units, then divide by number of ticks in 100 ns to get the nTimeout in ticks
		bInfTimeout = pDI->Timeout == 0L;
		if (!bInfTimeout)
			liMaxTicks = LiXDiv (LiNMul(pDI->Timeout,10000L), KeQueryTimeIncrement() );

		// Reset timout counter
		KeQueryTickCount (&liTicksStart) ;

		// Spin, reading C012 and busy waiting for Present bit
		for ( nRead = 0; nRead < nReadLength && !bTimeout; nRead++ ) {
			for (;;) {
				if ( READ_PORT_UCHAR(C012_isr) & C012_ISR_PRESENT ) {
					*pIOBuffer++ = READ_PORT_UCHAR (C012_idr);
					// Reset timout counter
					KeQueryTickCount (&liTicksStart);
					break;
				}
				else if (!bInfTimeout) {
					// Nothing Present, check timeout
					KeQueryTickCount (&liTicks);
					liTicks.QuadPart = liTicks.QuadPart - liTicksStart.QuadPart;
					if (liTicks.QuadPart > liMaxTicks.QuadPart) {
						bTimeout = TRUE;
						break;
					}
				}
			}// Wait for Present loop
		}// Read loop

	    // Read completed, last byte wasn't read if timeout occured       
	    if (bTimeout) nRead--;

    	pIrp->IoStatus.Information = nRead;		// Actual number read
	    pIrp->IoStatus.Status = STATUS_SUCCESS;

	}// Valid Read

    // Don't boost priority as completed with queing Irp
    IoCompleteRequest (pIrp, IO_NO_INCREMENT);

    return pIrp->IoStatus.Status;
}

