/*++

Authur: Version 1.1
	Michael Stevens
	michael@mech.eng.usyd.edu.au
Previous Author: Versions to 2.0
	Jari Salminen


Module Name:

    dispatch.c

Abstract:

    Device dispatch routing
    for the Transputer B008 Motherboard device driver.

Environment:

    Kernel mode

--*/

#include <ntddk.h>
#include <devioctl.h>
#include "B008drv.h"
#include "LinkIoctl.h"
#include "readwrite.h"
#include "isr.h"
#include "linksup.h"


NTSTATUS
TpdDispatch(
    IN    PDEVICE_OBJECT pDO,
    IN    PIRP pIrp
    )

/*++

Routine Description:
    This routine is the dispatch handler for the driver.  It is responsible
    for processing the IRPs.

Arguments:

    pDO - Pointer to device object.

    pIrp - Pointer to the current IRP.

Return Value:

    STATUS_SUCCESS if the IRP was processed successfully, otherwise an error
    indicating the reason for failure.

--*/

{
    PDEVICE_INFO		pDI;
    PIO_STACK_LOCATION	pIrpStack;
    NTSTATUS			Status;

	ULONG				IoCtlCode;
	PUCHAR				Port;
	ULONG				Timeout;

	PULONG				pIOBuffer;		// Pointer to transfer buffer (treated as an array of longs).
	ULONG				InBufferSize;	// Amount of data avail. from caller.
	ULONG				OutBufferSize;	// Max data that caller can accept.

    //  Initialize the irp info field.
    //      This is used to return the number of bytes transfered.

    pIrp->IoStatus.Information = 0;
    pDI = (PDEVICE_INFO)pDO->DeviceExtension;    // Get info struct

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

	IoCtlCode = pIrpStack->Parameters.DeviceIoControl.IoControlCode;

    //  Set default return status
    Status = STATUS_NOT_IMPLEMENTED;

    // Dispatch based on major fcn code.

    switch (pIrpStack->MajorFunction)
    {
	    case IRP_MJ_CREATE:
			Status = STATUS_SUCCESS;
		break;

		case IRP_MJ_CLOSE:
			Status = STATUS_SUCCESS;
		break;

    	case IRP_MJ_DEVICE_CONTROL: {
	        //  Dispatch on IOCTL
			Port = pDI->PortBase;
			Timeout = pDI->Timeout;

		    // Size of buffer containing data from application
		    InBufferSize  = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;

		    // Size of buffer for data to be sent to application
		    OutBufferSize = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;

		    pIrp->IoStatus.Information = OutBufferSize;		// Default

		    // NT copies inbuf here before entry and copies this to outbuf after
		    // return, for METHOD_BUFFERED IOCTL's.
		    pIOBuffer = (PULONG)pIrp->AssociatedIrp.SystemBuffer;

	        switch ( IoCtlCode ) {
				case IOCTL_LINK_RESET_LINK:
					B008ResetLink( Port );
			        Status = STATUS_SUCCESS;
				break;

				case IOCTL_LINK_ANALYSE_LINK:
					B008AnalyseLink( Port );
			        Status = STATUS_SUCCESS;
				break;

				case IOCTL_LINK_TEST_ERROR:
					if ( OutBufferSize == sizeof(ULONG) ) {
						*pIOBuffer = B008TestError(Port);
				        Status = STATUS_SUCCESS;
					} else {
				        Status = STATUS_INVALID_PARAMETER;
					}
				break;

				case IOCTL_LINK_TEST_READ:
					if ( OutBufferSize == sizeof(ULONG) ) {
						*pIOBuffer = B008TestRead(Port);
				        Status = STATUS_SUCCESS;
					} else {
				        Status = STATUS_INVALID_PARAMETER;
					}
				break;

				case IOCTL_LINK_TEST_WRITE:
					if ( OutBufferSize == sizeof(ULONG) ) {
						*pIOBuffer = B008TestWrite(Port);
				        Status = STATUS_SUCCESS;
					} else {
				        Status = STATUS_INVALID_PARAMETER;
					}
				break;

				case IOCTL_LINK_SET_TIMEOUT:
					if ( InBufferSize == sizeof(ULONG) ) {
						pDI->Timeout = *pIOBuffer;
				        Status = STATUS_SUCCESS;
					} else {
				        Status = STATUS_INVALID_PARAMETER;
					}
				break;

				case IOCTL_LINK_GET_TIMEOUT:
					if ( OutBufferSize == sizeof(ULONG) ) {
						*pIOBuffer = pDI->Timeout;
				        Status = STATUS_SUCCESS;
					} else {
				        Status = STATUS_INVALID_PARAMETER;
					}
				break;

				default:
					// Invalid IOCTL
					Status = STATUS_INVALID_PARAMETER;
			}// switch IOCTL

    	}// case MJ_DEVICE_CONTROL

    }// swicth Major function
    
    // We're done with I/O request.  Record the status of the I/O action.
    pIrp->IoStatus.Status = Status;

    // Don't boost priority when returning since this took little time.
    IoCompleteRequest( pIrp, IO_NO_INCREMENT );

    return Status;
}

