/*++

Authur: Version 1.1
	Michael Stevens
	michael@mech.eng.usyd.edu.au
Previous Author: Versions to 2.0
	Jari Salminen


Module Name:

    B008drv.c

Abstract:

    Driver initialisation and configuration
    for the Transputer B008 Motherboard device driver.

Environment:

    Kernel mode

--*/

#include <ntddk.h>
#include <devioctl.h>
#include "B008drv.h"
#include "LinkIoctl.h"
#include "dispatch.h"
#include "readwrite.h"
#include "isr.h"
#include "linksup.h"
#include "error.h"
#include "messages.h"
 

NTSTATUS
TpdCreateDevice(
    IN   PWSTR              PrototypeName,
    IN   DEVICE_TYPE        DeviceType,
    IN   PDRIVER_OBJECT     pDriverObject,
    OUT  PDEVICE_OBJECT     *ppDeviceObject,
	OUT  PUNICODE_STRING    pNTDeviceName,
	OUT  PUNICODE_STRING    pWin32DeviceName
    )

/*++

Routine Description:
    This routine creates the device object and provides the correct name
    for the symbolic link required between the device name and an entry
    in \DosDevices in order to allow Win32 applications to open the device.

    Ideally a name derived from a "Prototype", with a number appended at
    the end should be used.  For simplicity, just use the fixed name defined
    in the include file.  This means that only one device can be created.

	The device is created exclusive which follows the simple link model and
	makes life easy in the driver.


Arguments:

    PrototypeName - Name base, # WOULD be appended to this.
    DeviceType - Type of device to create
    pDriverObject - Pointer to driver object created by the system
    ppDeviceObject - Pointer to place to store pointer to created device object
	pWin32DeviceName - Pointer to place the WIN32 device name
	
Return Value:

    STATUS_SUCCESS if the device was created correctly, otherwise
    an error indicating the reason for failure.
    The device object pointer is set to NULL if not succesful.

--*/
{
    NTSTATUS Status;                        // Status of utility calls

    // Get UNICODE name for device.
    RtlInitUnicodeString(pNTDeviceName, PrototypeName);
    RtlInitUnicodeString(pWin32DeviceName, DOS_DEVICE_NAME);

	// Create it
    Status = IoCreateDevice(
                    pDriverObject,
                    sizeof(DEVICE_INFO),	// Device extension size
                    pNTDeviceName,
                    DeviceType,
                    0,						// No special characteristics flagged
                    TRUE,					// Exclusive
                    ppDeviceObject
                    );

    if (!NT_SUCCESS(Status)) {
        *ppDeviceObject = NULL;
    }

    return Status;
}


NTSTATUS
TpdGetParameters(
	IN   PUNICODE_STRING	RegistryPath,
	OUT  ULONG				*pBusPortBase,
    OUT  ULONG				*pPollRetry,
    OUT  ULONG				*pTimeout,
    OUT  ULONG				*pProcessHeader
    )
/*++
Routine Description:
	Retrieve the values from the Parameters key of our entry in the Registry.
	No defaults are supplied

Arguments:

Return Value:
	Status from Query operation

--*/
{
	NTSTATUS					Status = STATUS_INSUFFICIENT_RESOURCES;
    static  WCHAR               SubKeyString[] = L"\\Parameters";
    UNICODE_STRING              paramPath;
    RTL_QUERY_REGISTRY_TABLE    paramTable[5];

    // Compiled in DEFAULT Values;
	ULONG DEFAULT_PoleRetry = 100;
    ULONG DEFAULT_Timeout = 5000;
    ULONG DEFAULT_ProcessHeader = 0;
    
    //
    // Since the registry path parameter is a "counted" UNICODE string, it
    // might not be zero terminated.  For a very short time allocate memory
    // to hold the registry path as well as the Parameters key name zero
    // terminated so that we can use it to delve into the registry.
    //

    paramPath.MaximumLength = RegistryPath->Length + sizeof(SubKeyString);
    paramPath.Buffer = ExAllocatePool(PagedPool, paramPath.MaximumLength);

    if (paramPath.Buffer != NULL)
    {
        RtlMoveMemory(
            paramPath.Buffer, RegistryPath->Buffer, RegistryPath->Length);

        RtlMoveMemory(
            &paramPath.Buffer[RegistryPath->Length / 2], SubKeyString,
            sizeof(SubKeyString));

        paramPath.Length = paramPath.MaximumLength - 2;

		// Get registry info for Board
        RtlZeroMemory(&paramTable[0], sizeof(paramTable));

        paramTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
        paramTable[0].Name = L"BoardPortAddress";
        paramTable[0].EntryContext = pBusPortBase;
        paramTable[0].DefaultType = REG_DWORD;
        paramTable[0].DefaultData = NULL;
        paramTable[0].DefaultLength = sizeof(ULONG);

        paramTable[1].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[1].Name = L"PollRetry";
        paramTable[1].EntryContext = pPollRetry;
        paramTable[1].DefaultType = REG_DWORD;
        paramTable[1].DefaultData = &DEFAULT_PoleRetry;
        paramTable[1].DefaultLength = sizeof(ULONG);

        paramTable[2].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[2].Name = L"Timeout";
        paramTable[2].EntryContext = pTimeout;
        paramTable[2].DefaultType = REG_DWORD;
        paramTable[2].DefaultData = &DEFAULT_Timeout;
        paramTable[2].DefaultLength = sizeof(ULONG);

        paramTable[3].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[3].Name = L"ProcessHeader";
        paramTable[3].EntryContext = pProcessHeader;
        paramTable[3].DefaultType = REG_DWORD;
        paramTable[3].DefaultData = &DEFAULT_ProcessHeader;
        paramTable[3].DefaultLength = sizeof(ULONG);

        Status = RtlQueryRegistryValues(
            RTL_REGISTRY_ABSOLUTE,
            paramPath.Buffer, &paramTable[0], NULL, NULL);

        ExFreePool(paramPath.Buffer);
    }

	return Status;
}


NTSTATUS
TpdReportResources(
    IN   PDRIVER_OBJECT		DriverObject,
    IN   PDEVICE_OBJECT		DeviceObject,
    IN   PHYSICAL_ADDRESS	BusPortAddress,
    IN   ULONG				PortCount
    )
/*++
Routine Description:
	Register Resource (Ports) usage for Device.
Arguments:

Return Value:

	Combines the returned status from IoReportResourceUsage and
	any resource conflict as STATUS_DEVICE_CONFIGURATION_ERROR

--*/
{
    NTSTATUS			Status;
    /* Allocate the resource list.
       This already includes one CM_FULL_RESOURCE_DESCRIPTOR which in turn
       already include CM_PARTIAL_RESOURCE_LIST */

    UCHAR 				ResBuffer[sizeof(CM_RESOURCE_LIST)];
    PCM_RESOURCE_LIST	ResourceList = (PCM_RESOURCE_LIST)ResBuffer;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR ResourceDescriptor = ResourceList->List[0].PartialResourceList.PartialDescriptors;

    BOOLEAN				ResourceConflict;
	
    ResourceList->Count = 1;
    ResourceList->List[0].InterfaceType = Isa;
    ResourceList->List[0].BusNumber = 0;
    ResourceList->List[0].PartialResourceList.Count = 1;
    ResourceList->List[0].PartialResourceList.Version = 0;
    ResourceList->List[0].PartialResourceList.Revision = 0;

	// PORT resources
    ResourceDescriptor->Type = CmResourceTypePort;
    ResourceDescriptor->ShareDisposition = CmResourceShareDriverExclusive;
    ResourceDescriptor->Flags = CM_RESOURCE_PORT_IO;
    ResourceDescriptor->u.Port.Start = BusPortAddress;
    ResourceDescriptor->u.Port.Length = PortCount;
    ResourceDescriptor++;

    // Report resource usage and detect conflicts
    Status = IoReportResourceUsage(
                   NULL,	/* Default to "Other" */
                   DriverObject,
                   NULL,
                   0,
				   DeviceObject,
                   ResourceList,
                   (PUCHAR)ResourceDescriptor - (PUCHAR)ResourceList,
                   FALSE,
                   &ResourceConflict);

	// Log Error out own message reporting failure
    // Because resource errors haven't been built in IO_ERR such
    // values will not produce a device name.
    if (!NT_SUCCESS(Status)) {
      	TpdLogError( 
                  DeviceObject, 
                  0, 
                  0, 
                  0, 
                  1, 
                  Status, 
                  TPDLOG_RESOURCE_INVALID,
                  0,
                  NULL, 
                  0,
                  NULL
                 );
    };

	// Signal a ResourceConflict as an error, it will already be logged
    if (NT_SUCCESS(Status) && ResourceConflict)
        Status = STATUS_DEVICE_CONFIGURATION_ERROR;

    return Status;
}


void
TpdUnreportResources(
    IN   PDRIVER_OBJECT		DriverObject,
    IN   PDEVICE_OBJECT		DeviceObject
    )
/*++
Routine Description:
	UnRegister Resource for Device.
Arguments:

--*/
{
	NTSTATUS			Status;
	CM_RESOURCE_LIST	NullResourceList;
	BOOLEAN				ResourceConflict;
	
    // Unreport our resource usage, forget about the returned status
    RtlZeroMemory((PVOID)&NullResourceList, sizeof(NullResourceList) );
	Status = IoReportResourceUsage(
	   	               NULL,
        	           DriverObject,
            	       NULL,
                       0,
	                   DeviceObject,
    	               &NullResourceList,
        	           sizeof(NullResourceList),
            	       FALSE,
                	   &ResourceConflict);
	ASSERT (NT_SUCCESS(Status));
}


VOID
TpdUnload(
    PDRIVER_OBJECT DriverObject
    )

/*++

Routine Description:
    This routine prepares our driver to be unloaded.  It is responsible
    for freeing all resources allocated by DriverEntry as well as any
    allocated while the driver was running.  The symbolic link must be
    deleted as well.

Arguments:

    DriverObject - Pointer to driver object created by the system.

Return Value:

    None

--*/

{
	NTSTATUS Status;
    PDEVICE_INFO pDI;
    UNICODE_STRING Win32DeviceName;

    // Find our device information
    pDI = (PDEVICE_INFO)DriverObject->DeviceObject->DeviceExtension;
    ASSERT(pDI->DeviceType == LINK_TYPE);

    // Report we're not using any hardware.  If we don't do this
    // then we'll conflict with ourselves (!) on the next load
	TpdUnreportResources( DriverObject, DriverObject->DeviceObject );

	// Remove Device Object
	IoDeleteDevice( DriverObject->DeviceObject );

    // Assume all handles are closed down.
    // Delete the things we allocated - devices, symbolic links
    RtlInitUnicodeString(&Win32DeviceName, DOS_DEVICE_NAME);

    Status = IoDeleteSymbolicLink(&Win32DeviceName);
	ASSERT (NT_SUCCESS(Status));
}



NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath
    )

/*++

Routine Description:
    This routine is the entry point for the driver.  It is responsible
    for setting the dispatch entry points in the driver object and creating
    the device object.  Any resources such as ports, interrupts and DMA
    channels used must be reported.  A symbolic link must be created between
    the device name and an entry in \DosDevices in order to allow Win32
    applications to open the device.

Arguments:

    DriverObject - Pointer to driver object created by the system.

Return Value:

    STATUS_SUCCESS if the driver initialized correctly, otherwise an error
    indicating the reason for failure.

--*/

{
    ULONG				BusPortBase;			// Bus-relative Port location defined by user
    PHYSICAL_ADDRESS	BusPortAddress;			// NT Address form

    ULONG				PortCount = PORT_COUNT;	// Count of contiguous I/O ports
	PHYSICAL_ADDRESS	MappedPortAddress;		// HAL translated address
	ULONG				PortAddressSpace;		// Address space of ports
	PVOID				LogicalPortAddress;		// Actual logical address of port, maybe mapped on in IO space

	ULONG				InitialPollRetry;		// Initial values for device
	ULONG				InitialTimeout;
	ULONG				InitialProcessHeader;
	
    PDEVICE_OBJECT		pDeviceObject = NULL;	// Device Object, not created yet
    BOOLEAN				bDeviceResources = FALSE;	// Device has reported resources
    PDEVICE_INFO		pDI;  					// Device extension, information for each device.
	UNICODE_STRING		NTDeviceName;			// Device name created
	UNICODE_STRING      Win32DeviceName;		// Win32 name for the device

    NTSTATUS			Status;


	// Optional Debuging breakpoint
//	DbgBreakPoint();

	// Get boards configured parameters
	Status = TpdGetParameters( RegistryPath, &BusPortBase,
					&InitialPollRetry, &InitialTimeout, &InitialProcessHeader);

    if (!NT_SUCCESS(Status))
    {
        KdPrint( ("Tpd: Board Parameters could be retrived  %8X", Status) );

        return Status;
    }

	// Physical Address of Ports to register
    BusPortAddress.LowPart  = BusPortBase;
    BusPortAddress.HighPart = 0;

    // Get NT's Mapped Port Address and the address space its in
	//
    PortAddressSpace = 1;                        // Located in ISA IO space
    HalTranslateBusAddress( Isa,
                            0,
                            BusPortAddress,
                            &PortAddressSpace,
                            &MappedPortAddress );

    if (PortAddressSpace == 0)
    {    // Port is accessed through memory space - so get a logical address
         LogicalPortAddress = MmMapIoSpace(MappedPortAddress, PortCount, FALSE);
         if (LogicalPortAddress == NULL) {
         	Status = STATUS_SECTION_TOO_BIG;
	        KdPrint( ("Tpd: MmMapIoSpace %8X", Status) );
	        return Status;
         }
    }
    else
    {
         LogicalPortAddress = (PVOID)MappedPortAddress.LowPart;
    }

    // Create our Transputer device
	//
    Status = TpdCreateDevice(
                    TPD_DEVICE_NAME,
                    LINK_TYPE,
                    DriverObject,
                    &pDeviceObject,
                    &NTDeviceName,
                    &Win32DeviceName
                    );

    if (NT_SUCCESS(Status)) 
	{
		// Initialize the driver object dispatch table.
	    DriverObject->MajorFunction[IRP_MJ_CREATE]          = TpdDispatch;
    	DriverObject->MajorFunction[IRP_MJ_CLOSE]           = TpdDispatch;
		DriverObject->MajorFunction[IRP_MJ_READ]			= TpdRead;
		DriverObject->MajorFunction[IRP_MJ_WRITE]			= TpdWrite;
	    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = TpdDispatch;
		// Need to be able to unload if anything goes wrong
	    DriverObject->DriverUnload                          = TpdUnload;

	    // Set up the rest of the device info
	    // any alingment is good for us as we can do BYTE_ALIGNMENT
		pDeviceObject->Flags |= DO_BUFFERED_IO;

        // Initialize the device info for each device object.
        pDI = (PDEVICE_INFO)pDeviceObject->DeviceExtension;

		pDI->DeviceType			= LINK_TYPE;
        pDI->pDeviceObject		= pDeviceObject;
		pDI->PortAddressSpace	= PortAddressSpace;
        pDI->PortBase			= (PUCHAR)MappedPortAddress.LowPart;
		pDI->PollRetry			= InitialPollRetry;
		pDI->Timeout			= InitialTimeout;
		pDI->ProcessHeader		= InitialProcessHeader;

		// Device Create, create a 
	    Status = IoCreateSymbolicLink( &Win32DeviceName, &NTDeviceName );

		// Report resource usage, give up if there is a conflict
		//
		Status = TpdReportResources( DriverObject, pDeviceObject, BusPortAddress, PortCount);
	    if ( NT_SUCCESS(Status) )
    	{
			bDeviceResources = TRUE;

			// Disable the B008 interrupt system
			// Interrupt drivers in high impedance state
			WRITE_PORT_UCHAR( pDI->PortBase+OFFSET_OSR, 0 );

		}// if Resources reported

    }// if Device created


    // Any Error now requires release of resources and possibly the device.
    // Would need to add disconnection of interrupt if an error could occur after connection.
    //
    if (!NT_SUCCESS(Status))
    {
    	NTSTATUS cleanStatus;
        KdPrint( ("Tpd: Cleaning up after creation error %8X", Status) );

		if (bDeviceResources) {
    	    TpdUnreportResources(DriverObject, pDeviceObject);
		}
		
		if (pDeviceObject != NULL) {
			// Remove Device Object and its symbolic link
			IoDeleteDevice (pDeviceObject);
		    cleanStatus = IoDeleteSymbolicLink(&Win32DeviceName);
			ASSERT (NT_SUCCESS(cleanStatus));
		}
		
    }// if Error Cleanup

    return Status;
}

